use std::collections::hash_map::DefaultHasher;
use std::collections::{HashMap, HashSet};
use std::hash::Hash;
use std::time::Instant;
use itertools::Itertools;
use std::hash::Hasher;

enum Direction{
    North,
    South,
    East,
    West
}
pub fn execute_task01(content: &str) {
    let start = Instant::now();
    let gravity_1 = solve_01(content).unwrap();
    let duration = start.elapsed();

    assert_eq!(108955, gravity_1);
    println!("Day14 - Task01 - Duration: {duration:?} - One time gravity: {}", gravity_1)
}

pub fn execute_task02(content: &str) {
    let start = Instant::now();
    let gravity_1000000 = solve_02(content).unwrap();
    let duration = start.elapsed();

    assert_eq!(106689, gravity_1000000);
    println!("Day14 - Task02 - Duration: {duration:?} - 1 Million time gravity: {}", gravity_1000000)
}


fn solve_01(content: &str) -> anyhow::Result<usize> {
    let mut map = content
        .lines()
        .map(|chars| chars.chars().collect_vec())
        .collect_vec();

    move_direction(&mut map, Direction::North);

    Ok(map
        .iter()
        .rev()
        .enumerate()
        .map(|(i, row)| row.iter().filter(|c| c == &&'O').count() * (i+1))
        .sum())
}

fn solve_02(content: &str) -> anyhow::Result<usize> {
    let mut map = content
        .lines()
        .map(|chars| chars.chars().collect_vec())
        .collect_vec();

    let mut set: HashMap<u64, i32> = HashMap::new();
    let circles = 1_000_000_000;
    let mut from = 0;
    let mut to = 0;

    for i in 1..=circles {
        move_direction(&mut map, Direction::North);
        move_direction(&mut map, Direction::West);
        move_direction(&mut map, Direction::South);
        move_direction(&mut map, Direction::East);

        let hash = calculate_hash(&map);
        if let Some(step) = set.get(&hash) {
            from = step.clone();
            to = i;
            break;
        } else {
            set.insert(hash, i);
        }
    }

    let new_start = (circles - to) / (to - from) * (to -from);

    for i in new_start+from+1 ..=circles {
        move_direction(&mut map, Direction::North);
        move_direction(&mut map, Direction::West);
        move_direction(&mut map, Direction::South);
        move_direction(&mut map, Direction::East);
    }

    Ok(map
        .iter()
        .rev()
        .enumerate()
        .map(|(i, row)| row.iter().filter(|c| c == &&'O').count() * (i+1))
        .sum())
}

fn calculate_hash<T: Hash>(t: &T) -> u64 {
    let mut s = DefaultHasher::new();
    t.hash(&mut s);
    s.finish()
}

fn move_direction(map: &mut Vec<Vec<char>>, direction: Direction) {
    match direction {
        Direction::North => {
            let mut stones = vec![0; map[0].len()];

            for row in 0..map.len() {
                let width = map[row].len();
                for col in 0..width {
                    let char = map[row][col].clone();

                    match char {
                        'O' => {
                            if row > 0 && stones[col] != row {
                                map[stones[col]][col] = 'O';
                                map[row][col] = '.';
                                stones[col] += 1;
                            } else {
                                stones[col] = row +1;
                            }
                        },
                        '#' => stones[col] = row + 1,
                        '.' => (),
                        _ => unreachable!()
                    };
                }
            }
        }
        Direction::South => {
            let mut stones = vec![map.len() - 1; map[0].len()];

            for i in 0..map.len() {
                let row = map.len() -1 - i;
                let width = map[row].len();
                for col in 0..width {
                    let char = map[row][col].clone();

                    match char {
                        'O' => {
                            if stones[col] != row {
                                map[stones[col]][col] = 'O';
                                map[row][col] = '.';
                                stones[col] -= 1;
                            } else if row > 0 {
                                stones[col] = row - 1;
                            }
                        },
                        '#' => if row > 0 { stones[col] = row - 1 },
                        '.' => (),
                        _ => unreachable!()
                    };
                }
            }
        }
        Direction::West => {
            let width = map[0].len();
            let height = map.len();

            let mut stones = vec![0; height];

            for col in 0..width {
                for row in 0..height {
                    let char = map[row][col].clone();

                    match char {
                        'O' => {
                            if col > 0 && stones[row] != col {
                                map[row][stones[row]] = 'O';
                                map[row][col] = '.';
                                stones[row] += 1;
                            } else {
                                stones[row] = col + 1;
                            }
                        },
                        '#' => stones[row] = col + 1,
                        '.' => (),
                        _ => unreachable!()
                    };
                }
            }
        }
        Direction::East => {
            let width = map[0].len();
            let height = map.len();

            let mut stones = vec![width - 1; height];

            for i in 0..width {
                let col = width - 1 - i;
                for row in 0..height {
                    let char = map[row][col].clone();

                    match char {
                        'O' => {
                            if stones[row] != col {
                                map[row][stones[row]] = 'O';
                                map[row][col] = '.';
                                stones[row] -= 1;
                            } else if col > 0 {
                                stones[row] = col - 1;
                            }
                        },
                        '#' => if col > 0 { stones[row] = col - 1 },
                        '.' => (),
                        _ => unreachable!()
                    };
                }
            }
        }
    }
}


#[test]
fn test_solve_01() {
    let example = r#"O....#....
O.OO#....#
.....##...
OO.#O....O
.O.....O#.
O.#..O.#.#
..O..#O..O
.......O..
#....###..
#OO..#...."#;

    let result = solve_01(example).unwrap();

    assert_eq!(136, result);
}

#[test]
fn test_solve_02() {
    let example = r#"O....#....
O.OO#....#
.....##...
OO.#O....O
.O.....O#.
O.#..O.#.#
..O..#O..O
.......O..
#....###..
#OO..#...."#;

    let result = solve_02(example).unwrap();

    assert_eq!(64, result);
}
